import numpy as np
import matplotlib.pyplot as plt


# utilities


def simple_plot(x):
    """
    A simple plotting function, leaves some space on the top and bottom of the plot
    """

    maxy = np.max(x)
    miny = np.min(x)
    delta = maxy - miny

    plt.figure()
    plt.plot(x)
    plt.ylim( miny - delta*0.1, maxy + delta*0.1 )
    plt.show()


def DFT(x):
    return np.fft.fft(x, norm='ortho')


def IDFT(x):
    return np.real(np.fft.ifft(x, norm='ortho'))


def shift(x, N):
    return np.roll(x, N)


# GPS signal generation


def lfsr_sequence(N, taps):
    """
    Generate and return an 'N'-bit LFSR sequence of length 2^'N'-1, with the generating polynomial defined by 'taps'.
    >>> lfsr_sequence(4, [-4, -1])
    array([ 1,  1,  1,  1, -1,  1, -1,  1,  1, -1, -1,  1, -1, -1, -1])
    """

    length = 2**N-1

    y = np.zeros(length, dtype=np.int)
    x = np.ones(N, dtype=np.int)

    for i in range(length):
        y[i] = x[0]
        x[:-1], x[-1] = x[1:], -np.prod(x[taps])

    return y


# GPS G1 polynomial: X^10 + X^3 + 1
G1 = lfsr_sequence(10, [-3, -10])


# GPS G2 polynomial: X^10 + X^9 + X^8 + X^6 + X^3 + X^2 + 1
G2 = lfsr_sequence(10, [-2, -3, -6, -8, -9, -10])


def gps_sat_sequence(sat_id):
    """
    Generate and return the sequence for satellite with ID 'sat_id'
    """
    delay = [
        None,
        5,
        6,
        7,
        8,
        17,
        18,
        139,
        140,
        141,
        251,
        252,
        254,
        255,
        256,
        257,
        258,
        469,
        470,
        471,
        472,
        473,
        474,
        509,
        512,
        513,
        514,
        515,
        516,
        859,
        860,
        861,
        862,
    ]

    return -G1 * shift(G2, delay[sat_id])


def cross_correlation(x, y):
    """
    Return the cross correlation of x and y using DFT
    >>> cross_correlation(lfsr_sequence(4, [-4, -1]), lfsr_sequence(4, [-4, -1]))
    array([ 15.,  -1.,  -1.,  -1.,  -1.,  -1.,  -1.,  -1.,  -1.,  -1.,  -1.,
            -1.,  -1.,  -1.,  -1.])
    """
    x = shift(x[::-1], 1)
    return IDFT( np.sqrt(x.shape[0]) * DFT(x)*DFT(y) )


if __name__=='__main__':
    import doctest
    doctest.testmod()
